@@ -51,11 +51,12 @@ my_pty_fork(int *ptr_master_fd,
 
   if ((pid = fork()) < 0) {
     myerror("Cannot fork");
-    return(42); /* the compiler may not know that myeror() won't return */
+    return(42); /* the compiler may not know that myerror() won't return */
   } else if (pid == 0) {		/* child */
+    DEBUG_RANDOM_SLEEP;
     i_am_child = TRUE;		/* remember who I am */
-    unblock_signals(only_sigchld);	/* we have to unblock SIGCHLD because a subsequent exec() will retain the signal mask */
-    /* shouldn't we *restore* the signal mask? @@@ */
+    unblock_all_signals();
+    
     close(fdm);			/* fdm not used in child */
     ptytty_control_tty(fds, slave_name);
 
@@ -80,13 +81,14 @@ my_pty_fork(int *ptr_master_fd,
 
     return (0);
   } else {			/* parent */
-    install_signal_handlers();	/* by this time the child may be long dead! */
-    unblock_signals(only_sigchld); /* ... and that's why we had blocked the signals before forking */
-
+    srand(pid); DEBUG_RANDOM_SLEEP;
+    command_pid = pid;            /* the SIGCHLD signal handler needs this global variable */
+ 
     *ptr_master_fd = fdm;
-    if (!child_is_dead && tcgetattr(fdm, &pterm) < 0) {
+    if (!command_is_dead && tcgetattr(fdm, &pterm) < 0) {
       sleep(1);			/* we might be more succesful after the child command has
-				   initialized its terminal */
+				   initialized its terminal. As there is no reliable way to sense this
+				   from the parent, we just wait a little */
       if (tcgetattr(slave_pty_fd, &pterm) < 0) {
 	fprintf(stderr,		/* don't use mywarn() because of the strerror() message *within* the text */
 		"Warning: %s cannot determine terminal mode of %s\n"
@@ -107,54 +109,72 @@ my_pty_fork(int *ptr_master_fd,
       if (dup2(ttyfd, STDERR_FILENO) != STDERR_FILENO)
 	myerror("dup2 of stderr to ttyfd failed");
       close (ttyfd);
-    }	
-    return (pid);
+    }
+    if (renice)
+      nice(1);
+    return (pid); /* returns in parent and in child (and pid lets us determine who we are) */
   }
 }
 
-static int been_warned = 0;
-
 
 int
-slave_is_in_raw_mode()
+ slave_is_in_raw_mode()
 {
-  struct termios pterm;
-
+  struct termios *pterm_slave;
+  static int been_warned = 0;
+  int in_raw_mode;
   if (always_echo)
     return FALSE;
 
-  if (!child_is_dead && tcgetattr(slave_pty_fd, &pterm) < 0) {
-    /* race condition here: SIGCHLD may not yet heve been caught */
-    if (been_warned++ == 1)	/* only warn once, but not the first time (as this usually means that the rlwrapped command has just died) */
+  if (command_is_dead)
+    return FALSE; /* filter last words  too (even if ncurses-ridden) */
+  if (!(pterm_slave = my_tcgetattr(slave_pty_fd, "slave pty"))) {
+    if (been_warned++ == 1)	/* only warn once, but not the first time (as this usually means that the rlwrapped command has just died)
+				   - this is still a race when signals get delivered very late*/
       mywarn("tcgetattr error on slave pty (from parent process)");
-    return FALSE;
-  } else {
-    return (!(pterm.c_lflag & ICANON));
-  }
+    return TRUE;
+  }	
+ 
+  in_raw_mode = !(pterm_slave -> c_lflag & ICANON);
+  myfree(pterm_slave);
+  return in_raw_mode;
+  
 }
 
+
+
 void
 mirror_slaves_echo_mode()
 {				/* important e.g. when slave command asks for password  */
   struct termios *pterm_slave = NULL;
-  int should_echo_anyway = always_echo || always_readline;
+  int should_echo_anyway = always_echo || (always_readline && !dont_wrap_command_waits());
 
-  if (!child_is_dead && !always_echo && !(pterm_slave = get_pterm_slave()))
+  if ( !(pterm_slave = my_tcgetattr(slave_pty_fd, "slave pty")) ||
+      command_is_dead ||
+      always_echo 
+     )
     /* race condition here: SIGCHLD may not yet have been caught */
     return;
 
+  assert (pterm_slave != NULL);
+  
+  if (tcsetattr(STDIN_FILENO, TCSANOW, pterm_slave) < 0 && errno != ENOTTY) /* @@@ */
+     myerror ("cannot prepare terminal (tcsetattr error on stdin)");
 
+  term_eof = pterm_slave -> c_cc[VEOF];
+  
+  /* if the --always-readline option is set with argument "assword:", determine whether prompt ends with "assword:\s" */
   if (should_echo_anyway && password_prompt_search_string) {
     char *p, *q;
 
-    assert(strlen(saved_rl_state.prompt) < BUFFSIZE);
-    p = saved_rl_state.prompt + strlen(saved_rl_state.prompt) - 1;
+    assert(strlen(saved_rl_state.raw_prompt) < BUFFSIZE);
+    p = saved_rl_state.raw_prompt + strlen(saved_rl_state.raw_prompt) - 1;
     q =
       password_prompt_search_string + strlen(password_prompt_search_string) -
       1;
     while (*p == ' ')		/* skip trailing spaces in prompt */
       p--;
-    while (p >= saved_rl_state.prompt && q >= password_prompt_search_string)
+    while (p >= saved_rl_state.raw_prompt && q >= password_prompt_search_string)
       if (*p-- != *q--)
 	break;
 
@@ -163,7 +183,7 @@ mirror_slaves_echo_mode()
   }
 
 
-  if (!child_is_dead && (should_echo_anyway || pterm_slave->c_lflag & ECHO)) {
+  if (!command_is_dead && (should_echo_anyway || pterm_slave->c_lflag & ECHO)) {
     redisplay = TRUE;
   } else {
     redisplay = FALSE;
@@ -177,40 +197,32 @@ mirror_slaves_echo_mode()
 }
 
 
-struct termios *
-get_pterm_slave()
-{
-  struct termios *pterm_slave =
-    (struct termios *)mymalloc(sizeof(struct termios));
-  if (tcgetattr(slave_pty_fd, pterm_slave) < 0) {
-    if (been_warned++ == 1)	/* only warn once (but not the very first time, cf slave_is_in_raw_mode() above) */
-      mywarn("tcgetattr error on slave pty (from parent process)");
-    return NULL;
-  }
-  return pterm_slave;
-}
+
+
 
 
 void
 write_EOF_to_master_pty()
 {
-  struct termios *pterm_slave = get_pterm_slave();
+  struct termios *pterm_slave = my_tcgetattr(slave_pty_fd, "slave pty");
   char *sent_EOF = mysavestring("?");
 
-  *sent_EOF = pterm_slave->c_cc[VEOF];
+  *sent_EOF = (pterm_slave && pterm_slave->c_cc[VEOF]  ? pterm_slave->c_cc[VEOF] : 4) ; /*@@@ HL shouldn't we directly mysavestring(pterm_slave->c_cc[VEOF]) ??*/
+  DPRINTF1(DEBUG_TERMIO, "Sending %s", mangle_string_for_debug_log(sent_EOF, MANGLE_LENGTH));
   put_in_output_queue(sent_EOF);
-  free(pterm_slave);
+  myfree(pterm_slave);
   free(sent_EOF);
 }
 
 void
 write_EOL_to_master_pty(char *received_eol)
 {
-  struct termios *pterm_slave = get_pterm_slave();
+  struct termios *pterm_slave = my_tcgetattr(slave_pty_fd, "slave pty");
   char *sent_eol = mysavestring("?");
 
   *sent_eol = *received_eol;
-  switch (*received_eol) {
+  if (pterm_slave) { 
+    switch (*received_eol) {
     case '\n':
       if (pterm_slave->c_iflag & INLCR)
 	*sent_eol = '\r';
@@ -220,17 +232,81 @@ write_EOL_to_master_pty(char *received_eol)
 	return;
       if (pterm_slave->c_iflag & ICRNL)
 	*sent_eol = '\n';
-  }
+    }
+  }	
   put_in_output_queue(sent_eol);
-  free(pterm_slave);
+  myfree(pterm_slave);
   free(sent_eol);
 }
 
 void
 completely_mirror_slaves_terminal_settings()
 {
-  struct termios *pterm_slave = get_pterm_slave();
+  struct termios *pterm_slave = my_tcgetattr(slave_pty_fd, "slave pty");
+  log_terminal_settings(pterm_slave);
+  if (pterm_slave && tcsetattr(STDIN_FILENO, TCSANOW, pterm_slave) < 0 && errno != ENOTTY)
+    ;	/* myerror ("cannot prepare terminal (tcsetattr error on stdin)"); */
+  myfree(pterm_slave);
+}
 
-  if (tcsetattr(STDIN_FILENO, TCSANOW, pterm_slave) < 0 && errno != ENOTTY);	/* myerror ("cannot prepare terminal (tcsetattr error on stdin)"); */
-  free(pterm_slave);
+void
+completely_mirror_slaves_output_settings()
+{
+  
+  struct termios *pterm_stdin = my_tcgetattr(STDIN_FILENO, "stdin");
+  struct termios *pterm_slave = my_tcgetattr(slave_pty_fd, "slave pty");
+  if (pterm_slave && pterm_stdin) { /* no error message -  we can be called while slave is already dead */
+    pterm_stdin -> c_oflag = pterm_slave -> c_oflag;
+    tcsetattr(STDIN_FILENO, TCSANOW, pterm_stdin);
+  }	
+  myfree(pterm_slave);
+  myfree(pterm_stdin);
+}
+
+
+int dont_wrap_command_waits() {
+  static char command_wchan[MAXPATHLEN+1];
+  static int initialised = FALSE;
+  static int wchan_fd;
+  static int been_warned = 0;
+  char buffer[BUFFSIZE];
+  int nread;
+
+   if (!commands_children_not_wrapped)
+     return FALSE;
+   if (!initialised) { 	/* first time we're called after birth of child */
+    snprintf1(command_wchan, MAXPATHLEN , "/proc/%d/wchan", command_pid);
+    initialised =  TRUE;
+   }
+   if (command_is_dead)
+     return TRUE;  /* This is lazy!! signal may not have been delivered @@@ */
+   wchan_fd =  open(command_wchan, O_RDONLY);
+   if (wchan_fd < 0) { 
+     if (been_warned++ == 0) {
+       errno = 0; mywarn("you probably specified the -N (-no-children) option"
+			 "- but spying\non %s's wait status does not work on"
+			 " your system", command_name);
+     }
+     return FALSE;
+   }	
+   if (((nread = read(wchan_fd, buffer, BUFFSIZE -1)) > 0)) {
+     buffer[nread] =  '\0';
+     assert(!buffer[nread]); 
+     DPRINTF1(DEBUG_READLINE, "read commands wchan: <%s> ", buffer);
+     if (strstr(buffer, "wait")) /* @@@ HEY! does this always work? */
+       return TRUE;
+   }
+   close(wchan_fd);
+   return FALSE;
+}	
+
+
+int skip_rlwrap() { /* this function is called from sigTSTP signal handler. Is it re-entrant? */
+  if (dont_wrap_command_waits())
+    return TRUE;
+  if (always_readline)
+    return FALSE;
+  if (slave_is_in_raw_mode())
+    return TRUE;
+  return FALSE;
 }
