@@ -1,6 +1,7 @@
-/*  term.c: terminal handling, cursor movement etc. */
-
-/*  This program is free software; you can redistribute it and/or modify
+/*  term.c: terminal handling, cursor movement etc. 
+    (C) 2000-2007 Hans Lub
+    
+    This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License , or
     (at your option) any later version.
@@ -32,31 +33,29 @@ char *term_cursor_down;
 char *term_newline;
 
 int redisplay = 1;
+int newline_came_last = FALSE;
 
 struct termios saved_terminal_settings;	/* original terminal settings */
 int terminal_settings_saved = FALSE;	/*  saved_terminal_settings is valid */
-struct winsize window_size;	/* current window size */
+struct winsize winsize;	/* current window size */
 
 static char *term_cr;		/* carriage return (or 0, if none defined in terminfo) */
 static char *term_clear_line;
-static char *term_name;
+char *term_name;
 
-#ifdef HAVE_TERM_H
 static char *my_tgetstr (char *id) {
   char *term_string_buf = (char *)mymalloc(2048), *tb = term_string_buf;
   char *stringcap = tgetstr(id, &tb); /*  rl_get_termcap(id) should also get string capability but doesn't. Why? */
-  DPRINTF2(DEBUG_TERMIO, "tgetstr(\"%s\") = %s", id, (stringcap ? mangle_string_for_debug_log(stringcap,20) : "NULL"));
   char *retval = stringcap ? mysavestring(stringcap) : NULL; 
+  DPRINTF2(DEBUG_TERMIO, "tgetstr(\"%s\") = %s", id, (stringcap ? mangle_string_for_debug_log(stringcap,20) : "NULL"));
   free(term_string_buf);
   return retval;
 }
-#endif
 
 void
 init_terminal(void)
 {				/* save term settings and determine term type */
-  char *term_buf, *tb;
-  char *term_string_buf, *sb;
+  char *term_buf;
   int  tgetent_returnvalue;
   
   if (!isatty(STDIN_FILENO))
@@ -65,13 +64,17 @@ init_terminal(void)
     myerror("tcgetattr error on stdin");
   else
     terminal_settings_saved = TRUE;
-  if (ioctl(STDIN_FILENO, TIOCGWINSZ, &window_size) < 0)
+  if (ioctl(STDIN_FILENO, TIOCGWINSZ, &winsize) < 0)
     myerror("Could not get terminal size");
 
   /* init some variables: */
   term_name = getenv("TERM");
-  if (!term_name)
-    term_name = "dumb";
+  if (!term_name || strlen(term_name)==0) {
+    mywarn("environment variable TERM not set, assuming vt100\nSet it to your real terminal name if rlwrap misbehaves"); 
+    term_name = "vt100";
+  }
+ 
+  
   DPRINTF1(DEBUG_TERMIO, "TERM = %s", term_name);  
   term_buf = (char *)mymalloc(4096);
  
@@ -80,21 +83,26 @@ init_terminal(void)
   term_clear_line = NULL;
   term_cursor_hpos = NULL;
 
-#ifdef HAVE_TERM_H
   tgetent_returnvalue = tgetent(term_buf, term_name);
   if (tgetent_returnvalue > 0 ) {
     term_backspace 	= my_tgetstr("le");
     term_cr 		= my_tgetstr("cr");
-    term_clear_line 	= my_tgetstr("dl1");
+    term_clear_line 	= my_tgetstr("ce"); /* was: @@@ my_tgetstr("dl1") */
     term_clear_screen 	= my_tgetstr("cl");
     term_cursor_hpos 	= my_tgetstr("ch");
+    if (term_cursor_hpos && !strstr(term_cursor_hpos, "%p")) {
+      /* Some RedHat or Debian people with out-of sync devel packages will get problems when tgoto expects terminfo-style caps
+	 and their tgetstr returns termcap-style. This is a rather heavy-handed way of avoiding those problems: */    
+      DPRINTF0(DEBUG_TERMIO, "term_cursor_hpos appears to be termcap-style, not terminfo; NULLing it");
+      term_cursor_hpos = NULL;
+    }	
     term_cursor_up 	= my_tgetstr("up");
     term_cursor_down 	= my_tgetstr("do");
     term_newline 	= my_tgetstr("nl");							    
   } else {
     DPRINTF1(DEBUG_TERMIO, "tgetent returned %d", tgetent_returnvalue);
+    errno=0; mywarn("your $TERM is '%s' but %s couldn't find it in the terminfo database. Expect some problems.", term_name, program_name);
   }	
-#endif
 
   term_eof = saved_terminal_settings.c_cc[VEOF];
   term_stop = saved_terminal_settings.c_cc[VSTOP];
@@ -107,7 +115,7 @@ init_terminal(void)
 void
 set_echo(int yes)
 {
-  struct termios *pterm = get_pterm_slave();	/* mimic terminal settings of client */
+  struct termios *pterm = my_tcgetattr(slave_pty_fd, "slave pty");	/* mimic terminal settings of client */
 
   if (!pterm)			/* child has probably died */
     return;
@@ -118,15 +126,14 @@ set_echo(int yes)
     pterm->c_lflag |= ECHO;
   else				/* no */
     pterm->c_lflag &= ~ECHO;
+ 
+  log_terminal_settings(pterm);
   if (tcsetattr(STDIN_FILENO, TCSANOW, pterm) < 0 && errno != ENOTTY)
     ;	/* myerror ("cannot prepare terminal (tcsetattr error on stdin)"); */
   free(pterm);
 }
 
 
-#ifndef HAVE_TERM_H
-#define tputs(a,b,c)
-#endif
 
 void
 backspace(int count)
@@ -141,11 +148,16 @@ backspace(int count)
       my_putchar('\b');
 }
 
-void
+int
 cursor_hpos(int col)
 {
+  char *instantiated;
   assert(term_cursor_hpos != NULL);	/* caller has to make sure */
-  tputs(tgoto(term_cursor_hpos, 0, col), 1, my_putchar);
+  instantiated = tgoto(term_cursor_hpos, 0, col); /* tgoto with a command that takes one parameter: parameter goes to 2nd arg ("vertical position"). */
+  assert(instantiated);
+  DPRINTF2(DEBUG_TERMIO, "tgoto(term_cursor_hpos, 0, %d) = %s", col, mangle_string_for_debug_log(instantiated, 20));
+  tputs(instantiated, 1, my_putchar);
+  return TRUE;
 }
 
 void
@@ -165,25 +177,26 @@ clear_the_screen()
   if (term_clear_screen)
     tputs(term_clear_screen, 1, my_putchar);
   else
-    for (i = 0; i < window_size.ws_row; i++)
-      my_putchar('\n');		/* poor mans clear screen */
+    for (i = 0; i < winsize.ws_row; i++)
+      my_putchar('\n');		/* poor man's clear screen */
 }
 
 void
 clear_line()
 {
   int i;
-  int width = window_size.ws_col;
+  int width = winsize.ws_col;
   char *p, *spaces;
 
   cr();
   if (term_clear_line)
     tputs(term_clear_line, 1, my_putchar);
   else {
-    spaces = (char *)mymalloc(width);
+    spaces = (char *) mymalloc(width + 1);
     for (i = 0, p = spaces; i < width; i++, p++)
       *p = ' ';
-    write(STDOUT_FILENO, spaces, width);	/* poor mans clear line */
+    *p = '\0';
+    my_putstr(spaces);	/* poor man's clear line */
     free((void *)spaces);
   }
   cr();
@@ -193,7 +206,7 @@ clear_line()
 void
 curs_up()
 {
-  assert(term_cursor_up != NULL);	/* caller has to make sure */
+  /* assert(term_cursor_up != NULL);	caller has to make sure */
   tputs(term_cursor_up, 1, my_putchar);
 }
 
@@ -207,41 +220,57 @@ curs_down()
 int my_putchar(int c)
 {
   char ch = c;
-  ssize_t nwritten;
-
-  nwritten = write(STDOUT_FILENO, &ch, 1);
+  ssize_t nwritten = write(STDOUT_FILENO, &ch, 1);
   return (nwritten == -1 ? -1 : c);
 }
 
 
+void
+my_putstr(const char *string)
+{
+  int string_length = strlen(string);
+  DPRINTF2(DEBUG_TERMIO,"wrote %d bytes to stdout: %s", string_length,
+	   mangle_string_for_debug_log(string, MANGLE_LENGTH)); 
+  write_patiently(STDOUT_FILENO, string, string_length, "to stdout");
+  newline_came_last = (string[string_length - 1] == '\n'); /* remember whethere newline came last */
+
+}
+  
+
 static void test_termfunc(char *control_string, char *control_string_name, char* start, void (* termfunc)(), char *end)
 {
   char *mangled_control_string =
     (control_string ?
        add3strings("\"", mangle_string_for_debug_log(control_string,20),"\"")  :
        mysavestring("NULL"));
-  
   printf("\n%s = %s\n", control_string_name, mangled_control_string);
-  write(STDOUT_FILENO, start, strlen(start));
+  if (!control_string)
+    return;
+  my_putstr(start);
+  sleep(1);
   termfunc();
-  write(STDOUT_FILENO, end, strlen(end));
+  my_putstr(end);
+  sleep(1);
   free(mangled_control_string);
 }	
 
 static void backspace1 () { backspace(1); }
 static void cursor_hpos4 () { cursor_hpos(4);}
 
+
 void test_terminal()
 {
+  if (debug) {
+    debug_fp = fopen(DEBUG_FILENAME, "w");
+    setbuf(debug_fp, NULL);
+  }
   init_terminal();
-
   printf("\nTerminal is \"%s\"\n", term_name);  
   test_termfunc(term_backspace, "term_backspace", "This should print \"grape\": \ngras", &backspace1, "pe\n");
   test_termfunc(term_cr, "term_cr", "This should print \"lemon juice\": \napple", &cr, "lemon juice\n");
   test_termfunc(term_clear_line, "term_clear_line","This should print \"apple\": \npomegranate", &clear_line, "apple\n");
-  if (term_cursor_hpos)
-    test_termfunc(term_cursor_hpos, "term_cursor_hpos", "This should print \"pomegranate\": \npomelo", &cursor_hpos4, "granate\n");
-
+  test_termfunc(term_cursor_hpos, "term_cursor_hpos", "This should print \"pomegranate\": \npomelo", &cursor_hpos4, "granate\n");
+  test_termfunc(term_cursor_up, "term_cursor_up", "This should print \"blackberry\": \n blueberry\n", &curs_up,"black\n");
   printf("\n");
   
 }	
